#include "AsyncPipe.h"

#include <stdio.h>
#include <sys/time.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <unistd.h>
#include <errno.h>

AsyncPipe::AsyncPipe(const char* file, Type type, bool temp,
	QObject* parent, const char* name) : QObject(parent, name)
{
	m_fd = -1;
	m_type = type;
	m_temp = temp;
	m_file = QCString(file);
	//if(m_temp) ::mkfifo((const char*)m_file, DEFFILEMODE);
	::mkfifo((const char*)m_file, DEFFILEMODE);
	m_pTimer = new QTimer(this);
	m_pTimer->start(250, FALSE);
	connect(m_pTimer, SIGNAL(timeout()), this, SLOT(check()));
}

AsyncPipe::~AsyncPipe()
{
	if(m_fd >= 0) ::close(m_fd);
	if(m_temp) ::unlink((const char*)m_file);
}

void AsyncPipe::check()
{
	if(m_fd < 0){
		if(m_type == Read){
			m_fd = ::open((const char*)m_file, O_NONBLOCK|O_RDWR);
		} else {
			m_fd = ::open((const char*)m_file, O_NONBLOCK|O_WRONLY);
		}
	} else {
		int rc;
		struct timeval tv;
		tv.tv_sec = 0;
		tv.tv_usec = 0;
		fd_set fdset;
		fd_set fdset2;

		FD_ZERO(&fdset);
		FD_ZERO(&fdset2);

		FD_SET(m_fd, &fdset);
		FD_SET(m_fd, &fdset2);

		if(m_type == Read){
			rc = ::select(m_fd+1, &fdset, NULL, NULL, &tv);
		} else {
			rc = ::select(m_fd+1, &fdset2, &fdset, NULL, &tv);
			if(rc > 0 && FD_ISSET(m_fd, &fdset2)){
				::close(m_fd);
				m_fd = -1;
				emit unactivated();
				return;
			}
		}

		if(rc > 0 && FD_ISSET(m_fd, &fdset)){
			emit activated(m_fd);
		}
	}
}
