#include "CursorWidget.h"

#include <qapplication.h>
#include <qpainter.h>
#include <qimage.h>
#include <qbitmap.h>

#include <qpe/config.h>
#include <qpe/resource.h>

#include "cursor_data.xpm"
#include "marker_data.xpm"

CursorWidget::CursorWidget(QWidget* parent=0, const char* name=0)
		: QWidget(parent, name, WStyle_Customize | WStyle_Tool | WStyle_StaysOnTop)
{
	m_keep = FALSE;
	m_visible = FALSE;
	//setFocusPolicy(QWidget::NoFocus);
	m_current.setX(qApp->desktop()->width()/2);
	m_current.setY(qApp->desktop()->height()/2);
	m_pTimer = new QTimer(this);
	connect(m_pTimer, SIGNAL(timeout()),
		this, SLOT(hideCursor()));
	loadCursor();
	show();
}

void CursorWidget::loadCursorImage(const QString& data, const char* def[])
{
	QPixmap pix;
	if(!data.isEmpty()){
		pix = Resource::loadPixmap(data);
	}
	if(pix.isNull()){
		pix = QPixmap(def);
	}

	QImage maskImage;
	if(pix.mask()){
		maskImage = pix.mask()->convertToImage();
	} else {
		maskImage = pix.createHeuristicMask();
	}
	if(m_offsetX >= 0 && m_offsetY >= 0
		&& m_offsetX < maskImage.width() && m_offsetY < maskImage.height()){
		maskImage.setPixel(m_offsetX, m_offsetY, Qt::color0.pixel());
	}
	QBitmap mask;
	mask.convertFromImage(maskImage, QPixmap::MonoOnly);

	move(-QMAX(width(), pix.width()), -QMAX(height(), pix.height()));
	setFixedSize(pix.size());
	setBackgroundPixmap(pix);
	setMask(mask);
	update();
}

void CursorWidget::loadCursor()
{
	Config cfg("inputhelper");
	/* cursor data */
	cfg.setGroup("Cursor");
	const QString& data = cfg.readEntry("data");
	m_offsetX = cfg.readNumEntry("hotX", 7);
	m_offsetY = cfg.readNumEntry("hotY", 7);
	m_timeout = cfg.readNumEntry("timeout", 3000);

	/* cursor steps */
	cfg.setGroup("Step");

	m_steps[NORMAL] = cfg.readNumEntry("normal", 8);
	m_steps[SMALL] = cfg.readNumEntry("small", 2);
	m_steps[LARGE] = cfg.readNumEntry("large", 32);
	m_accels = cfg.readListEntry("accels", ',');

	m_wrapMode = cfg.readNumEntry("wrap", DEFAULT_WRAP);

	loadCursorImage(data, cursor_data);

#if 0
	raise();
	update();
#endif
}

void CursorWidget::loadMarker()
{
	Config cfg("inputhelper");
	/* marker data */
	cfg.setGroup("Marker");
	const QString& data = cfg.readEntry("data");
	m_offsetX = cfg.readNumEntry("hotX", 15);
	m_offsetY = cfg.readNumEntry("hotY", 15);
	m_timeout = cfg.readNumEntry("timeout", 500);

	loadCursorImage(data, marker_data);
}

void CursorWidget::showMarker()
{
	loadMarker();
	//setGeometry(m_current.x() - m_offsetX, m_current.y() -m_offsetY, width(), height());
	move(m_current.x() - m_offsetX, m_current.y() -m_offsetY);
	raise();
	QTimer::singleShot(m_timeout, this, SLOT(hideMarker()));
}

void CursorWidget::hideMarker()
{
	//setGeometry(-width(), -height(), width(), height());
	move(-width(), -height());
	loadCursor();
	if(m_visible){
		showCursor();
	} else {
		hideCursor();
	}
}

const QPoint& CursorWidget::showCursor()
{
	//setGeometry(m_current.x() - m_offsetX, m_current.y() -m_offsetY, width(), height());
	move(m_current.x() - m_offsetX, m_current.y() -m_offsetY);
	raise();
	m_visible = TRUE;
	//update();
	if(!m_keep) m_pTimer->start(m_timeout, TRUE);
	return m_current;
}

const QPoint& CursorWidget::showCursor(int x, int y, bool wrap)
{
	int width,height;
	m_current.setX(x);
	m_current.setY(y);
	width = qApp->desktop()->width();
	height = qApp->desktop()->height();
	if(m_wrapMode == NO_WRAP){
		wrap = FALSE;
	} else if(m_wrapMode == ALWAYS_WRAP){
		wrap = TRUE;
	}
	if(m_current.x() < 0){
		if(wrap){
			m_current.setX(width + m_current.x());
		} else {
			m_current.setX(0);
		}
	} else if(width < m_current.x()){
		if(wrap){
			m_current.setX(m_current.x() - width);
		} else {
			m_current.setX(width);
		}
	}
	if(m_current.y() < 0){
		if(wrap){
			m_current.setY(height + m_current.y());
		} else {
			m_current.setY(0);
		}
	} else if(height < m_current.y()){
		if(wrap){
			m_current.setY(m_current.y() - height);
		} else {
			m_current.setY(height);
		}
	} 
	return showCursor();
}

const QPoint& CursorWidget::moveCursor(int x, int y, bool wrap)
{
	return showCursor(m_current.x() + x, m_current.y() + y, wrap);
}

void CursorWidget::hideCursor()
{
	m_visible = FALSE;
	//setGeometry(-width(), -height(), width(), height());
	move(-width(), -height());
}

int CursorWidget::accelStep(int delta, int repeats)
{
	unsigned int i;
	for(i=0; i<m_accels.count(); i++){
		if(m_accels[i].toInt() > repeats) break;
	}
	return delta * (i+1);
}
