#ifndef _CURSOR_WIDGET_H_
#define _CURSOR_WIDGET_H_

#include <qwidget.h>
#include <qpixmap.h>
#include <qtimer.h>

class CursorWidget : public QWidget
{
	Q_OBJECT
public:
	enum Steps {
		NORMAL,
		SMALL,
		LARGE,
		STEP_COUNT,
	};
	enum WrapMode {
		NO_WRAP,
		DEFAULT_WRAP,
		ALWAYS_WRAP,
	};
	CursorWidget(QWidget* parent=0, const char* name=0);
	void loadCursor();
	const QPoint& showCursor(int x, int y, bool wrap=FALSE);
	const QPoint& moveCursor(int x, int y, bool wrap=FALSE);
	void keepCursor(bool keep=TRUE){
		m_keep = keep;
	}
	const QPoint& cursorPos(){
		return m_current;
	}
	int cursorStep(Steps kind){
		return m_steps[kind];
	}
	int accelStep(int delta, int repeat);

	void loadMarker();
public slots:
	const QPoint& showCursor();
	void hideCursor();
	void showMarker();
	void hideMarker();
protected:
private:
	void loadCursorImage(const QString& data, const char* def[]);

	int m_offsetX;
	int m_offsetY;
	QPoint m_current;

	int m_timeout;

	/* cursor step */
	int m_steps[STEP_COUNT];
	QStringList m_accels;

	bool m_keep;
	bool m_visible;
	QTimer* m_pTimer;
	int m_wrapMode;
private slots:
};

#endif /* _CURSOR_WIDGET_H_ */
