#include "InputHelperWidget.h"
#include "AsyncPipe.h"
#include "StringParser.h"
#include "KeyNames.h"
#include "QSafeDataStream.h"

#include <qcstring.h>
#include <qfile.h>
#include <qtextstream.h>
#include <qaccel.h>
#include <qclipboard.h>

#include <qpe/qpeapplication.h>
#include <qpe/qcopenvelope_qws.h>
#include <qpe/config.h>

#include <qwindowsystem_qws.h>
#include <qgfx_qws.h>
/* mouse support by Craig Graham */
#include <qsocketnotifier.h>
#include <fcntl.h>
#include <unistd.h>
#include <sys/stat.h>
#include <sys/types.h>

#define MOUSE_PIPE	"/var/spool/inputhelper/mouse"
#define KEY_PIPE	"/var/spool/inputhelper/key"
#define QCOP_PIPE	"/var/spool/inputhelper/qcop"

extern QScreen* qt_screen;

struct KeyEvent {
	KeyEvent(int u,int k,int m,int p,int r){
		unicode = u;keycode=k;modifiers=m;isPress=p;autoRepeat=r;
	}
	int unicode;
	int keycode;
	int modifiers;
	int isPress;
	int autoRepeat;
};

struct QCopData {
	QCopData(const QCString& m, const QByteArray& a){
		msg = m; data = a;
	}
	QCString msg;
	QByteArray data;
};

InputHelperWidget::InputHelperWidget(QWidget* parent, const char* name)
	: QWidget(parent, name)
{
	m_prevState = 0;
	//m_handle = FALSE;
	m_hook = NONE;
	m_sleep = FALSE;
	m_rest = -1;

	m_mouseFD = -1;
	m_mouseSN = NULL;

	m_saveEvents.setAutoDelete(TRUE);
	m_qcopQueue.setAutoDelete(TRUE);

	m_pTimer = new QTimer(this);
	connect(m_pTimer, SIGNAL(timeout()),
		this, SLOT(setHandler()));
	QCopChannel* qcopChannel = new QCopChannel("QPE/InputHelper", this);
	connect(qcopChannel, SIGNAL(received(const QCString&, const QByteArray&)),
		this, SLOT(receiveMessage(const QCString&, const QByteArray&)));

#if 0
	connect(qwsServer,
		SIGNAL(windowEvent(QWSWindow*, QWSServer::WindowEvent)),
		this,
		SLOT(windowEvent(QWSWindow*, QWSServer::WindowEvent)));
	qApp->installEventFilter(this);
#endif

	/* create mouse device */
	mkdir("/dev/input", 0777);
	mknod("/dev/input/mice", S_IFCHR|0666, (13 << 8) | 63);

	/* open mouse device */
	openMouse();

	initPipe();
	setHandler();
}

InputHelperWidget::~InputHelperWidget()
{
	closeMouse();
}

void InputHelperWidget::openMouse()
{
	if(m_mouseFD >= 0){
		closeMouse();
	}
	/* mouse support by Craig Graham */
	// Inintialise USB Mouse handling for SL6000
	m_mouseFD=open("/dev/input/mice",O_RDWR);
	m_mouseSN = new QSocketNotifier( m_mouseFD, QSocketNotifier::Read);
	if(m_mouseSN)
		connect( m_mouseSN, SIGNAL(activated(int)), SLOT(usbMouseDataAvailable(int)));
	m_usbIntellimouse=true;
	if(m_usbIntellimouse)
	{
		unsigned char mousedev_imex_seq[] = { 0xf3, 200, 0xf3, 200, 0xf3, 80 };
		unsigned char c;
		// set mouse to intellimouse mode for scroll wheel support
		write(m_mouseFD,mousedev_imex_seq,sizeof(mousedev_imex_seq));
		// read acknowledge...
		read(m_mouseFD,&c,1);
        }
	m_mouseByteCount=0;
	m_mouseLastMB=0;
}

void InputHelperWidget::closeMouse()
{
	if(m_mouseSN){
		disconnect(m_mouseSN, SIGNAL(activated(int)),
			this, SLOT(usbMouseDataAvailable(int)));
		delete m_mouseSN;
		m_mouseSN = NULL;
	}
	qApp->processEvents();
	if(m_mouseFD >= 0){
		::close(m_mouseFD);
		m_mouseFD = -1;
	}
}

/* mouse support by Craig Graham */
// SLOT for handling USB Mouse data on SL6000
void InputHelperWidget::usbMouseDataAvailable(int fd)
{
#define PS2MOUSE_LEFT   0x01
#define PS2MOUSE_RIGHT  0x02
#define PS2MOUSE_MIDDLE 0x04
#define PS2MOUSE_SIGN_X 0x10
#define PS2MOUSE_SIGN_Y 0x20
#define PS2MOUSE_OVERFLOW_X 0x40
#define PS2MOUSE_OVERFLOW_Y 0x80
	short dX,dY;
	read(fd,&m_mousePacket[m_mouseByteCount],1);
	if(!m_mouseByteCount)
	{
		if(m_mousePacket[m_mouseByteCount]&0x08) // try to sync up first byte....
			m_mouseByteCount++;
	}else{
		m_mouseByteCount++;
	}
	if(m_mouseByteCount==(m_usbIntellimouse?4:3))
	{
		int mb=0;
		dX=m_mousePacket[1];
		dY=m_mousePacket[2];
		if(m_mousePacket[0]&PS2MOUSE_SIGN_X)
			dX|=0xFF00;
		if(m_mousePacket[0]&PS2MOUSE_SIGN_Y)
			dY|=0xFF00;
		if(m_mousePacket[0]&PS2MOUSE_LEFT)
			mb|=Qt::LeftButton;
		if(m_mousePacket[0]&PS2MOUSE_RIGHT)
			mb|=Qt::RightButton;
		if(m_mousePacket[0]&PS2MOUSE_MIDDLE)
			mb|=Qt::MidButton;
		if(mb==m_mouseLastMB)
		{
			doMouse(MOVE, RELATIVE, dX, -dY, mb, FALSE);
		}else if(mb){
			// new buttons down...
			doMouse(PRESS, RELATIVE, dX, -dY, mb, FALSE);
		}else{
			doMouse(RELEASE, RELATIVE, dX, -dY, m_mouseLastMB, FALSE);
		}
		
		if(m_usbIntellimouse)
		{
			// handle scroll wheel stuff for Intellimouse mode
			short dZ;
			dZ=m_mousePacket[3]&0x07;
			if(m_mousePacket[3]&0x08)
				dZ|=0xFFF8;
			while(dZ<0)
			{
				QWSServer::sendKeyEvent(0xFFFF,Qt::Key_Up, 0, true, false);
				QWSServer::sendKeyEvent(0xFFFF,Qt::Key_Up, 0, false, false);
				dZ++;
			}
			while(dZ>0)
			{
				QWSServer::sendKeyEvent(0xFFFF,Qt::Key_Down, 0, true, false);
				QWSServer::sendKeyEvent(0xFFFF,Qt::Key_Down, 0, false, false);
				dZ--;
			}
		}
		
		m_mouseLastMB=mb;
		m_mouseByteCount=0;
	}
	
}


void InputHelperWidget::initPipe()
{
	m_mousePipe = new AsyncPipe(MOUSE_PIPE, AsyncPipe::Read, FALSE, this);
	m_keyPipe = new AsyncPipe(KEY_PIPE, AsyncPipe::Read, FALSE, this);
	m_qcopPipe = new AsyncPipe(QCOP_PIPE, AsyncPipe::Read, FALSE, this);
}

void InputHelperWidget::setHandler()
{
	m_sleep = FALSE;
	connect(m_mousePipe, SIGNAL(activated(int)), this, SLOT(readMouse(int)));
	connect(m_keyPipe, SIGNAL(activated(int)), this, SLOT(readKey(int)));
	connect(m_qcopPipe, SIGNAL(activated(int)), this, SLOT(readQCop(int)));
	while(QCopData* d=m_qcopQueue.dequeue()){
		QCopEnvelope e("QPE/InputHelper", d->msg);
		e.writeRawBytes(d->data.data(), d->data.size());
	}
}

void InputHelperWidget::sleepHandler(int msec)
{
	disconnect(m_mousePipe, SIGNAL(activated(int)), this, SLOT(readMouse(int)));
	disconnect(m_keyPipe, SIGNAL(activated(int)), this, SLOT(readKey(int)));
	disconnect(m_qcopPipe, SIGNAL(activated(int)), this, SLOT(readQCop(int)));
	m_pTimer->start(msec, TRUE);
	m_sleep = TRUE;
}

void InputHelperWidget::handleMouse(int unicode, int keycode, int modifiers,
	int isPress, int autoRepeat)
{
	qDebug("handleMouse[%x][%x][%x][%d][%d]", unicode, keycode, modifiers,
		isPress, autoRepeat);
	if(isPress){
		if(keycode == Qt::Key_Up || keycode == Qt::Key_Down
			|| keycode == Qt::Key_Right || keycode == Qt::Key_Left){
			int delta = m_cursor.cursorStep(CursorWidget::NORMAL);
			bool wrap=FALSE;
			if(!m_selecting){
				if(modifiers & Qt::ShiftButton){
					delta = m_cursor.cursorStep(CursorWidget::LARGE);
				}
				if(modifiers & Qt::ControlButton){
					wrap = TRUE;
					delta = m_cursor.cursorStep(CursorWidget::SMALL);
				}
			}
			if(autoRepeat){
				delta = m_cursor.accelStep(delta, m_repeats);
				m_repeats++;
			} else {
				m_repeats = 0;
			}
			switch(keycode){
			case Qt::Key_Up:
				doMouse(MOVE, RELATIVE, 0, -delta, Qt::LeftButton, wrap);
				break;
			case Qt::Key_Down:
				doMouse(MOVE, RELATIVE, 0, delta, Qt::LeftButton, wrap);
				break;
			case Qt::Key_Right:
				doMouse(MOVE, RELATIVE, delta, 0, Qt::LeftButton, wrap);
				break;
			case Qt::Key_Left:
				doMouse(MOVE, RELATIVE, -delta, 0, Qt::LeftButton, wrap);
				break;
			}
		} else {
			switch(keycode){
			case Qt::Key_Space:
				if(!autoRepeat){
					doMouse(PRESS, CURRENT, 0, 0, Qt::LeftButton);
					if(modifiers != 0) m_selecting = TRUE;
				}
				break;
			case Qt::Key_F23:
				if(!autoRepeat){
					if(modifiers == 0){
						doMouse(PRESS, CURRENT, 0, 0, Qt::RightButton);
					} else {
						doMouse(PRESS, CURRENT, 0, 0, Qt::MidButton);
					}
					if(modifiers != 0) m_selecting = TRUE;
				}
				break;
			case Qt::Key_Return:
				if(modifiers & Qt::ShiftButton){
					/* input Space */
					unicode = ' ';
					keycode = Qt::Key_Space;
					modifiers = 0;
				}
				goto SEND_KEY;
			case Qt::Key_Escape:
				doMouse(RELEASE);
				/* fall through */
			default:
SEND_KEY:
				QWSServer::sendKeyEvent(unicode, keycode, modifiers, isPress, autoRepeat);
				break;
			}
		}
	} else {
		bool released=FALSE;
		switch(keycode){
		case Qt::Key_Space:
			if(!autoRepeat){
				doMouse(RELEASE, CURRENT, 0, 0, Qt::LeftButton);
				released = TRUE;
			}
			break;
		case Qt::Key_F23:
			if(!autoRepeat){
				doMouse(RELEASE, CURRENT, 0, 0, Qt::RightButton);
				released = TRUE;
			}
			break;
		default:
			QWSServer::sendKeyEvent(unicode, keycode, modifiers, isPress, autoRepeat);
			break;
		}
		if(m_selecting && modifiers == 0) m_selecting = FALSE;
		if(released && m_rest > 0){
			if(--m_rest == 0){
				QCopEnvelope e("QPE/InputHelper", "handle()");
			}
		}
	}
}

void InputHelperWidget::doMouse(int action, int pos=CURRENT,
	int x=0, int y=0, int state=Qt::LeftButton, bool wrap=FALSE)
{
	qDebug("doMouse[%d][%d][%d][%d][%d]", action, pos, x, y, state);
	QPoint cursorPos;
	int button = (state & Qt::MouseButtonMask);

	switch(pos){
	case CURRENT:
		cursorPos = m_cursor.showCursor();
		break;
	case ABSOLUTE:
		cursorPos = m_cursor.showCursor(x, y, wrap);
		break;
	case RELATIVE:
		cursorPos = m_cursor.moveCursor(x, y, wrap);
		break;
	}
	if(action == MOVE){
		if(m_prevState){
			action = PRESS;
		} else {
			action = RELEASE;
		}
		state = m_prevState | (state & Qt::KeyButtonMask);
		button = 0;
	} else if(action == TOGGLE){
		if(m_prevState & button){
			action = RELEASE;
		} else {
			action = PRESS;
		}
	}
	//if(state) m_cursor.hideCursor();
	QPoint eventPos = qt_screen->mapToDevice(cursorPos, qApp->desktop()->size());
	if(action&PRESS){
		m_prevState |= button;
		//qDebug("sendMouseEvent([%d,%d],%d)", eventPos.x(), eventPos.y(), state);
		QWSServer::sendMouseEvent(eventPos, state);
	}
	if(action&RELEASE){
		m_prevState &= ~button;
		//qDebug("sendMouseEvent([%d,%d],%d)", eventPos.x(), eventPos.y(), 0);
		QWSServer::sendMouseEvent(eventPos, 0);
		//if(state) qApp->processEvents();
	}
	//if(state) m_cursor.showCursor();
}

bool InputHelperWidget::parseMouse(const QString& line)
{
	const QStringList& args = StringParser::split(' ', line);
	int pos = CURRENT;
	int state = Qt::LeftButton;
	int action = MOVE;
	int x=0, y=0;

	if(args.count() > 0){
		const QString& s = args[0].lower();
		if(s == "move"){
			action = MOVE;
		} else if(s == "press"){
			action = PRESS;
		} else if(s == "release"){
			action = RELEASE;
		} else if(s == "click"){
			action = CLICK;
		} else if(s == "toggle"){
			action = TOGGLE;
		} else if(s == "wait"){
			sleepHandler(args[1].toInt());
			return TRUE;
		}
	}
	if(args.count() > 2){
		if(args[1][0] == '+' || args[1][0] == '-'
			|| args[2][0] == '+' || args[2][0] == '-'){
			pos = RELATIVE;
		} else {
			pos = ABSOLUTE;
		}
		x = args[1].toInt();
		y = args[2].toInt();
	}
	if(args.count() > 3){
		const QString& s = args[3].lower();
		if(s == "left"){
			state = Qt::LeftButton;
		} else if(s == "right"){
			state = Qt::RightButton;
		} else if(s == "middle"){
			state = Qt::MidButton;
		}
	}
	if(args.count() > 4){
		const QString& s = args[4].lower();
		if(s.find("shift") >= 0){
			state |= Qt::ShiftButton;
		}
		if(s.find("ctrl") >= 0){
			state |= Qt::ControlButton;
		}
		if(s.find("alt") >= 0){
			state |= Qt::AltButton;
		}
	}
	doMouse(action, pos, x, y, state);
	return FALSE;
}

void InputHelperWidget::readMouse(int fd)
{
	QFile f;
	if(f.open(IO_ReadOnly, fd)){
		QTextStream stream(&f);
		QString line;
		while((line=stream.readLine()) != QString::null){
			if(parseMouse(line)) break;
		}
		f.close();
	}
}

void InputHelperWidget::doKey(int unicode, int keycode, int modifiers, int action,
	int repeat, int lowlevel)
{
	qDebug("doKey[%x][%x][%x][%d][%d][%d]", unicode, keycode, modifiers, action, repeat, lowlevel);
	if(action&PRESS){
		if(lowlevel){
			QWSServer::processKeyEvent(unicode, keycode, modifiers, TRUE, repeat);
		} else {
			QWSServer::sendKeyEvent(unicode, keycode, modifiers, TRUE, repeat);
		}
	}
	if(action&RELEASE){
		if(lowlevel){
			QWSServer::processKeyEvent(unicode, keycode, modifiers, FALSE, repeat);
		} else {
			QWSServer::sendKeyEvent(unicode, keycode, modifiers, FALSE, repeat);
		}
	}
}

void InputHelperWidget::getKeyCode(const QString& key,
	int& unicode, int& keycode, int& modifiers)
{
	modifiers = 0;
	int pos;
	QString keyname = key;
	pos = keyname.find("+SHIFT", 0, FALSE);
	if(pos > 0){
		modifiers |= Qt::ShiftButton;
		keyname.remove(pos, 6);
	}
	pos = keyname.find("+CTRL", 0, FALSE);
	if(pos > 0){
		modifiers |= Qt::ControlButton;
		keyname.remove(pos, 5);
	}
	pos = keyname.find("+ALT", 0, FALSE);
	if(pos > 0){
		modifiers |= Qt::AltButton;
		keyname.remove(pos, 4);
	}
	if(keyname.length() > 1){
		unicode = 0xffff;
		keycode = KeyNames::getCode(keyname);
		/* get unicode */
		const QWSServer::KeyMap* m;
		for(m=QWSServer::keyMap(); m->key_code != 0; m++){
			if(m->key_code == keycode){
				if(modifiers & Qt::ControlButton){
					unicode = m->ctrl_unicode;
				} else if(modifiers & Qt::ShiftButton){
					unicode = m->shift_unicode;
				} else {
					unicode = m->unicode;
				}
				break;
			}
		}
	} else if(modifiers & Qt::ControlButton){
		keycode = KeyNames::getCode(keyname.upper());
		unicode = 0xffff;
		const QWSServer::KeyMap* m;
		/* check keycode */
		for(m=QWSServer::keyMap(); m->key_code != 0; m++){
			if(m->key_code == keycode){
				unicode = m->ctrl_unicode;
				break;
			}
		}
	} else {
		const QWSServer::KeyMap* m;
		keycode = 0;
		unicode = keyname[0].unicode();
		/* check unicode */
		for(m=QWSServer::keyMap(); keycode == 0 && m->key_code != 0; m++){
			if(m->unicode == unicode){
				keycode = m->key_code;
				break;
			}
		}
		/* check shift_unicode */
		for(m=QWSServer::keyMap(); keycode == 0 && m->key_code != 0; m++){
			if(m->shift_unicode == unicode){
				keycode = m->key_code;
				modifiers |= Qt::ShiftButton;
				break;
			}
		}
		/* check ctrl_unicode */
		for(m=QWSServer::keyMap(); keycode == 0 && m->key_code != 0; m++){
			if(m->ctrl_unicode == unicode){
				keycode = m->key_code;
				modifiers |= Qt::ControlButton;
				break;
			}
		}
	}
}

bool InputHelperWidget::parseKey(const QString& line)
{
	const QStringList& args = StringParser::split(' ', line);
	int repeat = 0;
	int action = CLICK;

	if(args.count() > 0){
		int unicode, keycode, modifiers;
		unsigned int index=0;
		int lowlevel=0;
		if(args.count() > index){
			const QString& s = args[index].lower();
			if(s == "press" || s == "1"){
				action = PRESS;
			} else if(s == "release" || s == "0"){
				action = RELEASE;
			} else if(s == "click"){
				action = CLICK;
			} else if(s == "wait"){
				sleepHandler(args[index+1].toInt());
				return TRUE;
			}
		}
		index++;
		if(args[index].left(2) == "0x"){
			bool ok;
			unicode = args[index++].mid(2).toInt(&ok, 16);
			keycode = args[index++].mid(2).toInt(&ok, 16);
			modifiers = args[index++].mid(2).toInt(&ok, 16);
		} else {
			getKeyCode(args[index++], unicode, keycode, modifiers);
			if(keycode == Qt::Key_Shift){
				modifiers |= Qt::ShiftButton;
			} else if(keycode == Qt::Key_Control){
				modifiers |= Qt::ControlButton;
			} else if(keycode == Qt::Key_Alt){
				modifiers |= Qt::AltButton;
			}
		}
		if(args.count() > index){
			const QString& s = args[index].lower();
			if(s == "repeat" || s == "true" || s == "1" || s == "yes"){
				repeat = 1;
			}
		}
		index++;
		if(args.count() > index){
			const QString& s = args[index].lower();
			if(s == "low" || s == "true" || s == "1" || s == "yes"){
				lowlevel = 1;
			}
		}
		doKey(unicode, keycode, modifiers, action, repeat, lowlevel);
	}
	return FALSE;
}

void InputHelperWidget::readKey(int fd)
{
	QFile f;
	if(f.open(IO_ReadOnly, fd)){
		QTextStream stream(&f);
		QString line;
		while((line=stream.readLine()) != QString::null){
			if(parseKey(line)) break;
		}
		f.close();
	}
}

bool InputHelperWidget::parseQCop(const QString& line)
{
	const QStringList& args = StringParser::split(' ', line);

	if(args.count() > 1){
		if(args[0].lower() == "wait"){
			sleepHandler(args[1].toInt());
			return TRUE;
		}
		QCopEnvelope qcop(args[0].latin1(), args[1].latin1());
		QRegExp rxparam("[,()]");
		qDebug("qcop[%s][%s]", args[0].latin1(), args[1].latin1());
		const QStringList& params = QStringList::split(rxparam, args[1]);

		QString arg;
		for(uint i=1; i<params.count(); i++){
			if(args.count() > i+1){
				arg = args[i+1];
			} else {
				arg = QString::null;
			}
			qDebug("param[%s]", arg.latin1());
			if(params[i] == "QString"){
				qcop << arg;
			} else if(params[i] == "int"){
				qcop << arg.toInt();
			} else if(params[i] == "bool"){
				const QString& s = arg.lower();
				if(s == "true" || s == "1"){
					qcop << 1;
				} else if(s == "false" || s == "0"){
					qcop << 0;
				} else {
					qcop << 0;
				}
			} else if(params[i] == "QStringList"){
				const QStringList& listArg = StringParser::split(',', arg);
				qcop << listArg;
			}
		}
	}
	return FALSE;
}

void InputHelperWidget::readQCop(int fd)
{
	QFile f;
	if(f.open(IO_ReadOnly, fd)){
		QTextStream stream(&f);
		stream.setEncoding(QTextStream::UnicodeUTF8);
		QString line;
		while((line=stream.readLine()) != QString::null){
			if(parseQCop(line)) break;
		}
		f.close();
	}
}

void InputHelperWidget::receiveMessage(
		const QCString& msg, const QByteArray& data)
{
	//qDebug("message[%s]", (const char*)msg);
	if(m_sleep){
		m_qcopQueue.enqueue(new QCopData(msg, data));
		return;
	}
	QSafeDataStream stream(data, IO_ReadOnly);
	if(msg == "keyEvent(int,int,int,int,int)"){
		int unicode, keycode, modifiers, isPress, autoRepeat;
		stream >> unicode >> keycode >> modifiers >> isPress >> autoRepeat;
		if(m_hook == MOUSE){
			handleMouse(unicode, keycode, modifiers, isPress, autoRepeat);
		} else if(m_hook == KEY){
			recordKey(unicode, keycode, modifiers, isPress, autoRepeat);
		}
	} else if(msg == "mouse(QString)"){
		QString line;
		stream >> line;
		parseMouse(line);
	} else if(msg == "key(QString)"){
		QString line;
		stream >> line;
		parseKey(line);
	} else if(msg == "qcop(QString)"){
		QString line;
		stream >> line;
		parseQCop(line);
	} else if(msg == "clip(QString)"){
		QClipboard* cb = QApplication::clipboard();
		if(cb){
			QString text;
			stream >> text;
			cb->setText(text);
		}
	} else if(msg == "handle()"){
		if(m_hook == NONE){
			m_hook = MOUSE;
			m_selecting = FALSE;
			//m_handle = TRUE;
			m_cursor.loadCursor();
			m_cursor.keepCursor(TRUE);
			m_cursor.showCursor();
			if(m_rest != 0){
				QCopEnvelope e("QPE/KeyHelper", "hook(QString)");
				e << QString("QPE/InputHelper");
			}
		} else if(m_hook == MOUSE){
			m_hook = NONE;
			m_rest = -1;
			//m_handle = FALSE;
			m_cursor.keepCursor(FALSE);
			m_cursor.hideCursor();
			QCopEnvelope e("QPE/KeyHelper", "unhook()");
		}
	} else if(msg == "handle(int)"){
		stream >> m_rest;
		QCopEnvelope e("QPE/InputHelper", "handle()");
	} else if(msg == "position(QString)"){
		QString s;
		stream >> s;
		savePos(s);
	} else if(msg == "where()"){
		m_cursor.showMarker();
	} else if(msg == "record()"){
		if(m_hook == NONE){
			m_hook = KEY;
			m_recording = FALSE;
			m_saveEvents.clear();
			QCopEnvelope e("QPE/KeyHelper", "hook(QString)");
			e << QString("QPE/InputHelper");
		} else if(m_hook == KEY){
			m_hook = NONE;
			QCopEnvelope e("QPE/KeyHelper", "unhook()");
		}
	} else if(msg == "play()"){
		m_playFile = QString::null;
		QTimer::singleShot(0, this, SLOT(playKey()));
	} else if(msg == "play(QString)"){
		stream >> m_playFile;
		QTimer::singleShot(0, this, SLOT(playKey()));
	} else if(msg == "save(QString)"){
		QString s;
		stream >> s;
		saveKey(s);
	} else if(msg == "openMouse()"){
		openMouse();
	} else if(msg == "closeMouse()"){
		closeMouse();
	}
}

void InputHelperWidget::savePos(const QString &file)
{
	const QPoint& pos = m_cursor.cursorPos();
	QString position = QString::number(pos.x()) + " " + QString::number(pos.y());
	if(file.isEmpty()){
		QClipboard* cb = QApplication::clipboard();
		if(cb){
			cb->setText(position);
		}
	} else {
		QFile f(file);
		if(f.open(IO_WriteOnly|IO_Append)){
			QTextStream stream(&f);
			stream << position << "\n";
			f.close();
		}
	}
}

void InputHelperWidget::recordKey(int unicode, int keycode, int modifiers,
	int isPress, int autoRepeat)
{
	if(m_recording || isPress){
		m_saveEvents.append(new KeyEvent(unicode, keycode, modifiers, isPress, autoRepeat));
		m_recording = TRUE;
	}
	QWSServer::sendKeyEvent(unicode, keycode, modifiers, isPress, autoRepeat);
}

void InputHelperWidget::playKey()
{
	if(m_hook == KEY){
		m_hook = NONE;
		QCopEnvelope e("QPE/KeyHelper", "unhook()");
	}
	if(m_playFile.isNull()){
		QListIterator<KeyEvent> it(m_saveEvents);
		KeyEvent* ke;
		/* pressΤߤǽäƤ륤٥ȤϺ */
		for(it.toLast(); (ke=it.current())!=NULL; /*--it*/){
			if(ke->isPress){
				m_saveEvents.remove();
			} else {
				break;
			}
		}

		for(it.toFirst(); (ke=it.current())!=NULL; ++it){
			QWSServer::sendKeyEvent(ke->unicode, ke->keycode, ke->modifiers, ke->isPress, ke->autoRepeat);
		}
	} else {
		QFile f(m_playFile);
		QString s;
		QTextStream* stream = NULL;
		if(m_playFile.isEmpty()){
			QClipboard* cb = QApplication::clipboard();
			if(cb){
				s = cb->text();
				stream = new QTextStream(&s, IO_ReadOnly);
			}
		} else {
			if(f.open(IO_ReadOnly)){
				stream = new QTextStream(&f);
			}
		}
		if(stream){
			QString line;
			while((line=stream->readLine()) != QString::null){
				QCopEnvelope e("QPE/InputHelper", "key(QString)");
				e << line;
			}
			delete stream;
		}
	}
}

void InputHelperWidget::saveKey(const QString& file)
{
	if(m_saveEvents.isEmpty()) return;

	QFile f(file);
	QString s;
	QTextStream* stream = NULL;
	if(file.isEmpty()){
		stream = new QTextStream(&s, IO_WriteOnly);
	} else {
		if(f.open(IO_WriteOnly)){
			stream = new QTextStream(&f);
		}
	}
	if(stream){
		stream->flags(QTextStream::hex|QTextStream::showbase);
		QListIterator<KeyEvent> it(m_saveEvents);
		for(; it.current(); /*++it*/){
			KeyEvent* ke=it.current();
			++it;
			if(ke->isPress){
				if(it.current()){
					KeyEvent* ke2 = it.current();
					if(!ke2->isPress
						&& ke->unicode == ke2->unicode
						&& ke->keycode == ke2->keycode
						&& ke->modifiers == ke2->modifiers){
						ke->isPress = 2;
						++it;
					}
				} else {
					break;
				}
			}
			if(ke->isPress == 0){
				*stream << "release ";
			} else if(ke->isPress == 1){
				*stream << "press ";
			} else if(ke->isPress == 2){
				*stream << "click ";
			}
			const QString& name = KeyNames::getName(ke->keycode);
			if(name.isEmpty()){
				*stream << ke->unicode << " ";
				*stream << ke->keycode << " ";
				*stream << ke->modifiers;
			} else {
				if(name.length() == 1 &&
					(ke->modifiers == 0 || ke->modifiers == Qt::ShiftButton)){
					*stream << QChar(ke->unicode);
				} else {
					*stream << name;
				}
				if((ke->modifiers & Qt::ShiftButton) && ke->keycode != Qt::Key_Shift){
					*stream << "+SHIFT";
				}
				if((ke->modifiers & Qt::ControlButton) && ke->keycode != Qt::Key_Control){
					*stream << "+CTRL";
				}
				if((ke->modifiers & Qt::AltButton) && ke->keycode != Qt::Key_Alt){
					*stream << "+ALT";
				}
			}
			if(ke->autoRepeat){
				*stream << " repeat";
			}
			*stream << "\n";
		}
		if(!s.isEmpty()){
			QClipboard* cb = QApplication::clipboard();
			if(cb){
				cb->setText(s);
			}
		}
		delete stream;
	}

	m_saveEvents.clear();
}

#if 0
bool InputHelperWidget::eventFilter(QObject* o, QEvent* e)
{
#if 0
	switch(e->type()){
	case QEvent::Timer:
	case QEvent::MouseMove:
	case 12:
	case 13:
	case 14:
	case 70:
	case 72:
	case 26:
	case 27:
	case 17:
	case 37:
	case 71:
	case 50:
	case 11:
	case 10:
	case 32:
	case 6:
	case 7:
	case 2:
	case 3:
	case 4:
		break;
	case QEvent::ShowMaximized:
		m_cursor.raise();
	default:
		qDebug("####################################eventFilter[%d]", e->type());
	}
#else
#if 0
	if(e->type() == QEvent::FocusIn){
		QTimer::singleShot(0, &m_cursor, SLOT(raise()));
	}
	if(e->type() == QEvent::ShowMaximized){
		QTimer::singleShot(0, &m_cursor, SLOT(raise()));
	}
#endif
#endif
	return QWidget::eventFilter(o, e);
}
#endif

#if 0
void InputHelperWidget::windowEvent(QWSWindow* w, QWSServer::WindowEvent e)
{
#if 0
	//if(w->client()->identity() != NULL/* && w->name() != NULL*/){
	//if(w->client()->identity() != NULL){
		switch(e){
		//case QWSServer::Create:
		//case QWSServer::Raise:
		case QWSServer::Show:
			qDebug("raise()[%x]", e);
			//m_cursor.raise();
			QTimer::singleShot(0, &m_cursor, SLOT(raise()));
			break;
		default:
			break;
		}
	//}
#else
	if(e == QWSServer::Show) QTimer::singleShot(0, &m_cursor, SLOT(raise()));
#endif
}
#endif