#ifndef _INPUTHELPER_WIDGET_H_
#define _INPUTHELPER_WIDGET_H_

#include <qwidget.h>
#include <qlist.h>
#include <qtimer.h>
#include <qqueue.h>
#include <qwindowsystem_qws.h>
#include "CursorWidget.h"

struct KeyEvent;
struct QCopData;
class AsyncPipe;

class InputHelperWidget : public QWidget
{
	Q_OBJECT
public:
	InputHelperWidget(QWidget* parent=0, const char* name=0);
	~InputHelperWidget();

	enum Position{
		CURRENT,
		ABSOLUTE,
		RELATIVE,
	};
	enum Action{
		MOVE	= 0x0000,
		PRESS	= 0x0001,
		RELEASE	= 0x0002,
		CLICK	= (PRESS|RELEASE),
		TOGGLE	= 0x0004,
	};
	enum Hook {
		NONE,
		MOUSE,
		KEY,
	};
	void doMouse(int action, int pos=CURRENT, int x=0, int y=0,
		int state=Qt::LeftButton, bool wrap=FALSE);
	void doKey(int unicode, int keycode, int modifiers, int action,
		int repeat, int lowlevel);
protected:
	//bool eventFilter(QObject* o, QEvent* e);
protected slots:
	//void windowEvent(QWSWindow* w, QWSServer::WindowEvent e);
	/* mouse support by Craig Graham */
	void usbMouseDataAvailable(int fd);
private:
	void initPipe();

	void getKeyCode(const QString& key, int& unicode, int& keycode, int& modifiers);

	bool parseMouse(const QString& line);
	bool parseKey(const QString& line);
	bool parseQCop(const QString& line);

	void handleMouse(int unicode, int keycode, int modifiers, int isPress, int autoRepeat);
	void recordKey(int unicode, int keycode, int modifiers, int isPress, int autoRepeat);
	void saveKey(const QString& file);
	void savePos(const QString &file);
	void sleepHandler(int msec);



	CursorWidget m_cursor;
	int m_prevState;
	Hook m_hook;
	QList<KeyEvent> m_saveEvents;
	QQueue<QCopData> m_qcopQueue;
	bool m_sleep;
	bool m_recording;
	bool m_selecting;
	int m_repeats;
	int m_rest;
	QString m_playFile;
	QTimer* m_pTimer;
	AsyncPipe* m_mousePipe;
	AsyncPipe* m_keyPipe;
	AsyncPipe* m_qcopPipe;
	
	/* mouse support by Craig Graham */
	// USB Mouse Support for SL6000
	QSocketNotifier *m_mouseSN;
	int m_mouseFD;
	int m_mouseByteCount;
	unsigned char m_mousePacket[4];
	int m_mouseLastMB;
	bool m_usbIntellimouse;
	
private slots:

	void openMouse();
	void closeMouse();

	void setHandler();
	void playKey();
	void readMouse(int fd);
	void readKey(int fd);
	void readQCop(int fd);
	void receiveMessage(const QCString& msg, const QByteArray& data);
};

#endif /* _INPUTHELPER_WIDGET_H_ */
